/* SIP extension for IP connection tracking.
 *
 * (C) 2006 
 * based on RR's ip_conntrack_sip.c and other modules.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/udp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_sip.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("sercomm suzhou <sercomm@sdc.sercomm.com>");
MODULE_DESCRIPTION("SIP connection tracking helper");

static DECLARE_LOCK(sipbf_lock);
struct module *ip_conntrack_sip = THIS_MODULE;

#define MAX_PORTS	8
static int ports[MAX_PORTS];
static int ports_c;
#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

static unsigned int sip_timeout = SIP_TIMEOUT;

#ifdef MODULE_PARM
MODULE_PARM(sip_timeout,  "i");
#endif


int ct_sip_get_info(const char *dptr, size_t dlen, 
				unsigned int *matchoff, 
				unsigned int *matchlen, 
				struct sip_header_nfo *hnfo,
				u_int16_t *port);
EXPORT_SYMBOL(ct_sip_get_info);

#if 0 
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#if 0
#define DEBUGP_G printk
#else
#define DEBUGP_G(format, args...)
#endif

static int header_len(const char *dptr, const char *limit, int *shift, u_int16_t *port);
static int digits_len(const char *dptr, const char *limit, int *shift, u_int16_t *port);
static int epaddr_len(const char *dptr, const char *limit, int *shift, u_int16_t *port);
static int skp_digits_len(const char *dptr, const char *limit, int *shift, u_int16_t *port);
static int skp_epaddr_len(const char *dptr, const char *limit, int *shift, u_int16_t *port);

struct sip_header_nfo ct_sip_hdrs[] = {
	{ 	/* Via header */
		"Via:",		sizeof("Via:") - 1,
		"\r\nv:",	sizeof("\r\nv:") - 1, /* rfc3261 "\r\n" */
		"UDP ", 	sizeof("UDP ") - 1,
		epaddr_len
	},
	{ 	/* Contact header */
		"Contact:",	sizeof("Contact:") - 1,
		"\r\nm:",	sizeof("\r\nm:") - 1,
		"sip:",		sizeof("sip:") - 1,
		skp_epaddr_len
	},
	{ 	/* Content length header */
		"Content-Length:", sizeof("Content-Length:") - 1,
		"\r\nl:",	sizeof("\r\nl:") - 1,
		":",		sizeof(":") - 1, 
		skp_digits_len
	},
	{	/* SDP media info */
		"\nm=",		sizeof("\nm=") - 1,	
		"\rm=",		sizeof("\rm=") - 1,
		"audio ",	sizeof("audio ") - 1,
		digits_len
	},
	{ 	/* SDP owner address*/	
		"\no=",		sizeof("\no=") - 1, 
		"\ro=",		sizeof("\ro=") - 1,
		"IN IP4 ",	sizeof("IN IP4 ") - 1,
		epaddr_len
	},
	{ 	/* SDP connection info */
		"\nc=",		sizeof("\nc=") - 1, 
		"\rc=",		sizeof("\rc=") - 1,
		"IN IP4 ",	sizeof("IN IP4 ") - 1,
		epaddr_len
	},
	{ 	/* Requests headers */
		"sip:",		sizeof("sip:") - 1,
		"sip:",		sizeof("sip:") - 1, /* yes, i know.. ;) */
		"@", 		sizeof("@") - 1, 
		epaddr_len
	},
	{ 	/* SDP version header */
		"\nv=",		sizeof("\nv=") - 1,
		"\rv=",		sizeof("\rv=") - 1,
		"=", 		sizeof("=") - 1, 
		digits_len
	},
	{ 	/* Warning header req_src_ip*/
		"Warning:", sizeof("Warning:") - 1,
		"Warning:",	sizeof("Warning:") - 1,
		"req_src_ip=",		sizeof("req_src_ip=") - 1, 
		epaddr_len
	},
	{ 	/* Warning header req_src_port*/
		"Warning:", sizeof("Warning:") - 1,
		"Warning:",	sizeof("Warning:") - 1,
		"req_src_port=",		sizeof("req_src_port=") - 1, 
		digits_len
	},
	{   /* Call-ID header*/
		"Call-ID:",	sizeof("Call-ID:") - 1,
		"Call-ID:",	sizeof("Call-ID:") - 1,
		": ",		sizeof(": ") - 1,
		header_len
	}
};
EXPORT_SYMBOL(ct_sip_hdrs);


static int header_len(const char *dptr, const char *limit, int *shift, u_int16_t *port)
{
	int len = 0;
	DEBUGP_G("call-id : dptr = %p, %c,%c,%c", dptr, *dptr, *(dptr+1), *(dptr+2));
    while ((dptr <= limit) && (*dptr == '\r' || *dptr == '\n'))
            dptr++;

    while (dptr <= limit) 
    {
            if (*dptr == '\r' || *dptr == '\n')
                    break;
            dptr++;
            len++;
    }
    return len;
}

static int digits_len(const char *dptr, const char *limit, int *shift, u_int16_t *port)
{
	int len = 0;	
	while (dptr <= limit && isdigit(*dptr)) 
	{
		dptr++;
		len++;
	}
	return len;
} 

/* get digits lenght, skiping blank spaces. */
static int skp_digits_len(const char *dptr, const char *limit, int *shift, u_int16_t *port)
{
	for (; dptr <= limit && *dptr == ' '; dptr++)
		(*shift)++;
		
	return digits_len(dptr, limit, shift, port);
}

/* Simple ipaddr parser.. */
static int parse_ipaddr(const char *cp,	const char **endp, 
			u_int32_t *ipaddr, const char *limit)
{
	unsigned long int val;
	int i, digit = 0;
	
	for (i = 0, *ipaddr = 0; cp <= limit && i < 4; i++) 
	{
		digit = 0;
		if (!isdigit(*cp))
			break;
		
		val = simple_strtoul(cp, (char **)&cp, 10);
		if (val > 0xFF)
			return -1;
	
		((uint8_t *)ipaddr)[i] = val;	
		digit = 1;
	
		if (*cp != '.')
			break;
		cp++;
	}
	if (!digit)
		return -1;
	
	if (endp)
		*endp = cp;

	return 0;
}

/* skip ip address. returns it lenght. */
static int epaddr_len(const char *dptr, const char *limit, int *shift, u_int16_t *port)
{
	const char *aux = dptr;
	u_int32_t ip;
	
	*port = 0;
	
	if (parse_ipaddr(dptr, &dptr, &ip, limit) < 0) 
	{
		DEBUGP("ip: %s parse failed.!\n", dptr);
		return 0;
	}

	/* Port number */
	if (*dptr == ':') 
	{
		dptr++;
		*port = simple_strtoul(dptr, NULL, 10);
		dptr += digits_len(dptr, limit, shift, port);
	}
	return dptr - aux;
}

/* get address lenght, skiping user info. */
static int skp_epaddr_len(const char *dptr, const char *limit, int *shift, u_int16_t *port)
{
	for (; dptr <= limit && *dptr != '@'; dptr++)
		(*shift)++;	
	
	if (*dptr == '@') 
	{
		dptr++;
		(*shift)++;
		return epaddr_len(dptr, limit, shift, port);
	}	
	return  0;
}

/* Returns 0 if not found, -1 error parsing. */
int ct_sip_get_info(const char *dptr, size_t dlen, 
		unsigned int *matchoff, 
		unsigned int *matchlen,
		struct sip_header_nfo *hnfo,
		u_int16_t *port)
{
	const char *limit, *aux, *k = dptr;
	int shift = 0;
	
	limit = dptr + (dlen - hnfo->lnlen);

	while (dptr <= limit) 
	{
		if ((strncmp(dptr, hnfo->lname, hnfo->lnlen) != 0) &&
			(strncmp(dptr, hnfo->sname, hnfo->snlen) != 0))
		{
			dptr++;
			continue;
		}
		aux = ct_sip_search(hnfo->ln_str, dptr, hnfo->ln_strlen, 
						ct_sip_lnlen(dptr, limit));
		if (!aux) 
		{
			DEBUGP("'%s' not found in '%s'.\n", hnfo->ln_str, hnfo->lname);
			return -1;
		}
		aux += hnfo->ln_strlen;
		
		*matchlen = hnfo->match_len(aux, limit, &shift, port);
		if (!*matchlen)
			return -1;

		*matchoff = (aux - k) + shift; 
		
		DEBUGP_G("%s match succeeded! - len: %u\n", hnfo->lname, *matchlen);
		return 1;
	}
	DEBUGP("%s header not found.\n", hnfo->lname);
	return 0;
}

static int set_expected_rtp(
			struct ip_conntrack *ct,
			enum ip_conntrack_info ctinfo, 
			u_int32_t ipaddr, u_int16_t port,
			const char *dptr,
			unsigned char flag_port,
			unsigned char current_flag,
			u_int16_t ret_via_port)
{
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_sip_expect *exp_sip_info = &exp->help.exp_sip_info;

	memset(&expect, 0, sizeof(expect));

	exp->tuple = ((struct ip_conntrack_tuple)
		{ { ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip, { 0 } },
		  { ipaddr, { .udp = { htons(port) } }, IPPROTO_UDP }});
		  
	exp->mask = ((struct ip_conntrack_tuple) 
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { .udp = { 0xFFFF } }, 0xFF }});

	exp->expectfn = NULL;
	exp_sip_info->port = port;
	exp_sip_info->ipaddr_sip=ipaddr;
	exp_sip_info->flag_port_contant = flag_port;
	exp_sip_info->current_flag = current_flag;
	exp_sip_info->via_port = ret_via_port;
	exp_sip_info->contact_port = port;
	
	exp->master = ct;
	
	ip_conntrack_expect_related(ct, &expect);
	DEBUGP("set_expected_rtp flag_port_contant=0x%2x,current_flag=0x%2x,via_port=%d,port = %d\n",flag_port,current_flag,ret_via_port,port); 
	return NF_ACCEPT;
}

static int sip_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	unsigned int dataoff;
	int ret = NF_ACCEPT;
	int matchoff, matchlen;
	u_int32_t ipaddr,ipaddr_contact,ipaddr_via;
	u_int16_t port=0,ret_port=0,ret_contact_port=0,ret_via_port=0,src_port=0;
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *dptr = (const char *)udph + sizeof(struct udphdr);
	unsigned int udplen = len - iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	int i;
	unsigned char flag_port = 0;
	u_int32_t ip;

	DEBUGP_G("gordon enter sip_help\n");	

	src_port = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
	/* No Data ? */ 
	dataoff = iph->ihl*4 + sizeof(struct udphdr);
	DEBUGP_G("**************header:"); 
 
	for(i=0; i< 10; i++)
		DEBUGP_G("%c",*(dptr+i)); 
	 	DEBUGP_G("************\nbefore ip_ct_refresh\n");       

	ip_ct_refresh(ct, sip_timeout * HZ);
	
	DEBUGP_G("after ip_ct_refresh\n");    
	DEBUGP_G("return accept, do nothing in conntrack sip\n");   

	return NF_ACCEPT;
	
	
	LOCK_BH(&sipbf_lock);
	
	if ((ctinfo) < IP_CT_IS_REPLY) {
		ip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		port = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
	} else {
		ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
		port = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port;
	}
	
	
	
	if ((ctinfo) >= IP_CT_IS_REPLY)
		goto out;

	/* After this point NAT, could have mangled skb, so 
	   we need to recalculate payload lenght. */

	if (datalen < (sizeof("SIP/2.0 200") - 1))
		goto out;
	
	/* Get ip and port address from SIP Contact header*/
	if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
	    &ct_sip_hdrs[POS_CONTACT], &ret_contact_port) > 0) 
	{
		ipaddr_contact = 0;
		if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr_contact, 
		    dptr + datalen) < 0) 
		{
			ret = NF_DROP;
			goto out;
		}
		if (0 != ret_contact_port )
		{
			flag_port |= CONTACT_FLAG;	
			DEBUGP("this packet's contact field  include port %d,contact ip= %u.%u.%u.%u\n",ret_contact_port,NIPQUAD(ipaddr_contact)); 
		}

	}
	
	/* Get ip and port address from SIP Via header*/
	if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
	    &ct_sip_hdrs[POS_VIA], &ret_via_port) > 0) 
	{
		ipaddr_via = 0;

		if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr_via, 
		    dptr + datalen) < 0) 
		{
			ret = NF_DROP;
			goto out;
		}
			DEBUGP("via field  src_ip=%u.%u.%u.%u , via contant ip = %u.%u.%u.%u\n",NIPQUAD(ip),NIPQUAD(ipaddr_via)); 
				
		if(0 == ret_via_port && ip == ipaddr_via)
		{
			DEBUGP("this packet's via field  without via port\n",ret_via_port); 
		}
		else
		{
			flag_port |= VIA_FLAG;
			if(ip == ipaddr_via)
			DEBUGP("this packet's via field  include port %d\n",ret_via_port); 
			else
			DEBUGP("this packet's via field  include port %d, but it's not via port\n",ret_via_port); 
		}
	}

	DEBUGP("this packet's contact field  port=%d,udp packet src port=%d, via field port=%d\n",ret_contact_port,ntohs(src_port),ret_via_port); 

	if(0 != ret_contact_port && ret_contact_port != ntohs(src_port) && ip == ipaddr_via)
	{
			flag_port |= CONTACT_IN_EXP;
	DEBUGP("this packet's contact field  port=%d but udp packet src port=%d\n",ret_contact_port,ntohs(src_port)); 
	}

	/*add expectation for contact port*/
	if (0 != ret_contact_port )
	{
		if (ret_contact_port < 1024) 
		{
			ret = NF_DROP;
			goto out;
		}

		ret = set_expected_rtp(ct, ctinfo,
					ipaddr_contact, ret_contact_port, dptr, flag_port, CONTACT_FLAG, ret_via_port);
	}				



	/* RTP info only in some SDP pkts */
	if (memcmp(dptr, "INVITE", sizeof("INVITE") - 1) != 0 && 
	    memcmp(dptr, "SIP/2.0 200", sizeof("SIP/2.0 200") - 1) != 0) 
	{
		goto out;
	}

	/* Get ip and port address from SDP packet. */
	if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
	    &ct_sip_hdrs[POS_CONECTION], &ret_port) > 0) 
	{

		/* We'll drop only if there are parse problems. */
		if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr, 
		    dptr + datalen) < 0) 
		{
			ret = NF_DROP;
			goto out;
		}
		if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
		    &ct_sip_hdrs[POS_MEDIA], &ret_port) > 0) 
		{

			port = simple_strtoul(dptr + matchoff, NULL, 10);
			if (port < 1024) 
			{
				ret = NF_DROP;
				goto out;
			}

			ret = set_expected_rtp(ct, ctinfo,
						ipaddr, port, dptr, flag_port, MEDIA_FLAG, ret_via_port);
		}
	}


	out:
		UNLOCK_BH(&sipbf_lock);
		return ret;
}

static struct ip_conntrack_helper sip[MAX_PORTS];
static char sip_names[MAX_PORTS][10];

static void __exit fini(void)
{
	int i = 0;
	for (; i < ports_c; i++) 
	{
		DEBUGP("unregistering helper for port %d\n", ports[i]);
		ip_conntrack_helper_unregister(&sip[i]);
	} 
}

static void  fini_err(void)
{
	int i = 0;
	for (; i < ports_c; i++) 
	{
		DEBUGP("unregistering helper for port %d\n", ports[i]);
		ip_conntrack_helper_unregister(&sip[i]);
	} 
}

static int __init init(void)
{
	int i, ret;
	char *tmpname;

	if (ports_c == 0)
		ports[ports_c] = SIP_PORT;

		ports[1] = 5060;
//	for (i = 0; i < ports_c; i++) {
	for (i = 0; i < MAX_PORTS && ports[i]; i++) 
	{		/* Create helper structure */
		memset(&sip[i], 0, sizeof(struct ip_conntrack_helper));

		sip[i].tuple.dst.protonum = IPPROTO_UDP;
		sip[i].tuple.src.u.udp.port = htons(ports[i]);
		sip[i].mask.src.u.udp.port = 0xFFFF;
		sip[i].mask.dst.protonum = 0xFF;
		sip[i].max_expected = 16;
		sip[i].timeout = 3 * 60; /* 3 minutes */
		sip[i].me = THIS_MODULE;
		sip[i].help = sip_help;

		tmpname = &sip_names[i][0];
		if (ports[i] == SIP_PORT)
			sprintf(tmpname, "sip");
		else
			sprintf(tmpname, "sip-%d", i);
		sip[i].name = tmpname;

		DEBUGP("port #%d: %d\n", i, ports[i]);

		ret=ip_conntrack_helper_register(&sip[i]);
		if (ret) 
		{
			DEBUGP("ERROR registering helper for port %d\n",
				ports[i]);
			fini_err();
			return(ret);
		}
		ports_c++;
	}
	return(0);
}
#ifdef CONFIG_IP_NF_NAT_NEEDED
EXPORT_SYMBOL(sipbf_lock);
#endif

module_init(init);
module_exit(fini);
